;************************************************************************************************
;*                                                                                              *
;* Druckluftsteuerung mit ATMega48                                                              *
;*                                                                                              *
;* Ein Khlschrankaggregat pumpt Druckluft in ein Luftkessel. Der Druck wird in einem           *
;* eistellbaren Bereich (0..7 Bar) gehalten. Dazu wird ein Drucksensor benutzt , der ber den   *
;* A/D Wandler eingelesen wird und den Kompressor je nach vorhandenem Druck Ein/Ausschaltet.    *
;* Der vorhandene Druck kann jederzeit ber eine 3stellige 7Segmentanzeige abgelesen werden.    *
;* Die untere und die obere Schaltschwelle kann ber 3 Taster eingegeben werden und wird dann   *
;* im EEprom des AVR gespeichert.                                                               * 
;*                                                                                              *
;* Hardware:                                                                                    *
;* Drucksensor, MPX5700, 0-7 Bar, an PC3 angeschlossen laut Datenblatt (51k, 50pF Ausgang - GND)*
;* 3 Taster PC0..2, gegen GND, interne Pullups                                                  *
;* 3 7Segmentanzeigen, gem. Anode, a..f,DP an PortD, Anoden an PB0..2                           *
;* Relais Kompressor an PB3, Relais Magnetventil an PB4                                         *
;*                                                                                              *
;* Der ATMega 48 wird mit 8MHz (interner RC-Oszillator) betrieben                               *
;*                                                                                              *
;* Autor: J.Woetzel (juergen@avr-projekte.de)                                                   *
;* Version 1.00 vom 07/11/10                                                                    *
;*                                                                                              *
;************************************************************************************************
.include "m48def.inc"

.def	null	=	r0	;Konstante
.def	taste0	=	r1	;Tasten
.def	taste1	=	r2
.def	taste2	=	r3
.def	seg0	=	r4	;7Seg. Digits
.def	seg1	=	r5
.def	seg2	=	r6
.def	s		=	r7	;S-Register
.def	druckmin=	r8	;Variable fr Druck..
.def	druckmax=	r9
.def	ventil	=	r10	;.. Magnetventil

.def	temp1	=	r16         
.def	temp2	=	r17       
.def	temp3	=	r18       
.def	temp4	=	r19     
.def	tick	=	r20	;Timervariable, wird in der ISR runtergezhlt
.def	repeat	=	r21	;Tastenwiderholung
.def	flags	=	r22	;diverse Flags, siehe unten

;Flag-Register Bits
.equ	enter	=	0		;Tasten
.equ	plus	=	1
.equ	minus	=	2
;.equ	min		=	3		;Druck
.equ	max		=	4
.equ	DPunkt	=	7		;Anzeige


.equ	rep_kurz=	40		;Tastenwiderholung 40*2mS
.equ	rep_lang=	250		;erster Tastendruck, nach 250*2 mS Repeat ein

.org 0
rjmp	res

.org OVF0addr
	 
ISR:in		s,sreg			;Statusbits merken
	push	temp1
	push	temp2
	dec		tick			;Zhler fr wait
;Display multiplexen. Jeden Interupt (alle 2 mS) eine Stelle weiterschalten (gem. Anode) und den entsprechenden 7 Segmentcode
;in seg0, seg1, seg2 auf PortB ausgeben.

	in		temp2,portb		;Aktuelles Digit holen und
	andi	temp2,0b111		;ausmaskieren
	in		temp1,portb		;Display dunkel (negative Logik)
	sbr		temp1,0b111		;Alle 3 gem. Anoden abschalten
	out		portb,temp1

_mux0:
	cpi		temp2,0b011		;War die zuletzt gezeigte die 3. Stelle ?(negative Logik / gem. Anode)		
	brne	_mux1			;wenn nicht, Stelle 2 prfen
	cbi		portb,0			;Stelle 1 Einschalten
	out		portd,seg0		;Saft auf die Kathoden
	tst		flags			;Dezimalpunkt ?
	brpl	exit0
	ldi		temp1,$7f		;Maske Bit7 lschen
	and		seg0,temp1		;DP ein
	rjmp	exit0
_mux1:
	cpi		temp2,0b110		;War die zuletzt gezeigte die 1. Stelle ?
	brne	_mux2
	cbi		portb,1
	out		portd,seg1
	rjmp	exit0
_mux2:
	cbi		portb,2			;kann nur noch Stelle 3 sein
	out		portd,seg2

;------------------------------------------------------------------------------------------------------------
;3 Taster (Enter, Plus und Minus, PinB 3..5) entprellen
;Alle 2 mS wird (in diesem Fall) der entspr. Taster  in das Register <taste(n)> geschoben.
;Ist <taste> = 0, war der Taster die letzten 8 Interrupts gedrckt(im Main einfach tst taste).
;Ist das Carry zustzlich gesetzt, wird in einem Flagregister das entspr. Flankenbit gesetzt, das
;im Hauptprogramm nach Verarbeitung des Tastendruck wieder gelscht wird.
;So ist sichergestellt, das jeder Tastendruck nur einmal verarbeitet wird
;------------------------------------------------------------------------------------------------------------
exit0:
	in		temp1,pinc

;Enter, ohne Repeat

	lsr		temp1     	  	;(Enter-Taste) ins Carry
	rol		taste0   	   	;und in Taste0 schieben
	brcc	ex1     	  	;Es muss eine 1 (nicht gedrckt) und danach
	brne	ex1     	 	;8 mal eine 0 (gedrckt) folgen, erst dann ist der Tastendruck gltig.          
	sbr		flags,1<<enter 	;Flag (Taste gedrckt, Negative Flanke, Carry+Taste = 100000000) in Bit0 setzen..
           			  		;.. wird in der Main, nach Auswertung wieder gelscht.
;Plus und Minus, mit Repeat

ex1:
	lsr		temp1			;(Minus-Taste) ins Carry
	rol		taste1			;und in Taste1 schieben
	brne	ex2				;Taste ist 0 wenn die letzten 8 ISR gedrckt war			
	dec		repeat			;Repeatzhler-1
	brne	rp0				;Das Carry enthlt immer noch den wert von Pinc vor 9 ISRs
	ldi		repeat,rep_kurz	;Repeatwert neu laden 40*2mS=0,08 Sek.
	sec						;Taste durch Repeat gedrckt
rp0:brcc	ex2				;...wie oben
	sbr		flags,1<<minus	

ex2:
	lsr		temp1			;(Plus-Taste) ins Carry .....
	rol		taste2			;wie oben
	brne	ex3		
	dec		repeat
	brne	rp1
	ldi		repeat,rep_kurz	
	sec
rp1:brcc	ex3	
	sbr		flags,1<<plus	;Bit2 setzen

ex3:tst		taste1			;Wenn keine Taste gedrckt ist Repeatwert (1. Tastendruck = lang) laden
	breq	ex4
	tst		taste2
	breq	ex4
	ldi		repeat,rep_lang	;250*2mS =ca. 0,5 Sek
;---------------------------------------------------------------------------------------------------------
;Nach dem Abschalten des Aggregats wird Ventil (in der Main) mit 250 geladen. Hier wird
;PortB 4 ca. 0,5 Sekunden auf Lo gezogen.
ex4:
	tst		ventil			;Zeit abgelaufen ?
	breq	ex5				;Wenn ja, Sprung
	cbi		portb,4			;Portb4 auf 0
	dec		ventil			;Zeit-1 (250*2mS)
	brne	ex5				;Wenn noch was da ist, auf Lo lassen
	sbi		portb,4			;Fertig, Rckschlagventil aus

;------------------------------------------------------------------------------------------------------------
;ISR Ende
ex5:
	pop		temp2
	pop		temp1
	out		sreg,s			;Statusbits zurcklesen
	reti					;in der Main weitermachen
;***************************************************************************************************************

res:
    ldi     temp1, LOW(RAMEND)		; Stackpointer initialisieren
    out     SPL, temp1
    ldi     temp1, HIGH(RAMEND)
    out     SPH, temp1

	ldi		temp1,7					;Tasten 1..3
	out		portc,temp1				;Pullups ein

 	ldi		temp1,0xff				;PortD auf Ausgang (7Seg a..f,DP)
	out		ddrd,temp1
	out		portd,temp1

 	ldi		temp1,0x1f				;PortB auf Ausgang
	out		ddrb,temp1				;0..2 = 7-Seg. gem Anoden, 3 = Relais
	out		portb,temp1				;PortB Display erstmal dunkel,Relais aus

 	clr		null					;hehe
	ldi		flags,1<<DPunkt|1<<max	;DP ein, max. Druck
	clr		ventil
	rcall	load					;Zuletzt eingestellten Druck aus EEprom laden
	sei								;ISR Los

; ADC initialisieren: ADC0, Vcc als Referenz, Single Conversion, Vorteiler 128
 
    ldi     temp1, (1<<REFS0) |(1<<ADLAR)|1<<mux0|1<<mux1          ; Kanal 3, int. Ref.=5V, Analogwerte Linksbndig
    sts     ADMUX, temp1
    ldi     temp1, (1<<ADEN) | (1<<ADPS0) | (1<<ADPS1)| (1<<ADPS2)
    sts     ADCSRA, temp1

	ldi		temp1,1<<cs00|1<<cs01	;TC0/64, ISR ca. alle 2mS (8MHz/256/64)
	out		TCCR0B,temp1	
	ldi		temp1,1<<TOIE0			;Timer0 overflow interrupt
	sts		TIMSK0,temp1
;------------------------------------------------------------------------------------------------------
loop:
	clr		temp4
	clr		temp2
	clr		temp3
sample_adc:
 	ldi		temp1,1<<adsc|1<<aden
    sts     ADCSRA,temp1        ;ADC Los
wait_adc:
	lds		temp1,adcsra
	andi	temp1,1<<adsc
	brne	wait_adc
    lds		temp1, ADCL        	; immer zuerst LOW Byte lesen
    lds		temp1, ADCH      	; danach das mittlerweile gesperrte High Byte
	add		temp2,temp1			;256 Messwerte addieren
	adc		temp3,null			;Carry dazu
	dec		temp4
	brne	sample_adc			
	rcall	print_z				;in Temp3 steht nun der gltige (8Bit) Wert

;-----------------------------------------------------------------
;Relais schalten

	sbrc	flags,max		;Voll machen ?
	rjmp	cpmax			;wenn ja, Sprung

;War gerade Voll, warten bis Druck unteren Schaltpunkt erreicht hat 

cpmin:
	cp		temp3,druckmin	;Sind wir schon unter minimum ?
	brcc	mx0				;wenn nein, Sprung (Relais aus)
	cbi		portb,3			;Relais einschalten
	sbr		flags,1<<max	;Voll machen
	rjmp	relx			;Relais Routine verlassen
mx0:
	sbi		portb,3			;Relais aus
	rjmp	relx			;Relais Routine verlassen
	
cpmax:
	cp		temp3,druckmax	;Max. Druck erreicht ?
	brcc	mx1				;wenn ja Srung (Relais aus)
	cbi		portb,3			;Relais ein
	sbr		flags,1<<max	;Flag setzten(Voll machen)
	rjmp	relx			;Relais Routine verlassen
mx1:
	sbi		portb,3			;Relais aus
	cbr		flags,1<<max	;Flag lschen, warten bis Druck unteren Schaltpunkt erreicht hat
	ldi		temp1,250
	mov		ventil,temp1	;Rckschlagventil (0,5 sec.) starten (der Rest macht die ISR)
relx:
	sbrc	flags,enter		;Enter Taste ?
	rjmp	menu

	rcall	wait			;kurz warten wg. Display
	rjmp	loop

;----------------------------------------------------------------------------------------------

menu:
	sbi		portb,3						;Relais aus
	cbr		flags,1<<enter|1<<DPunkt	;Flags Entertaste, Dezimalpunkt lschen
	rcall	scroll_low					;scrollt "SET Lo" ber das Display
	sbr		flags,1<<dpunkt
	mov		temp3,druckmin				;Wert, unterer Schaltpunkt -> Display	
	rcall	print_z
mlop:
	sbrc	flags,enter					;Tasten abfragen
	rjmp	next						;bernehmen und Sprung -> Set high (Maximaldruck)
	sbrc	flags,plus					;+ Taste
	rjmp	up							;Wert erhhen
	sbrc	flags,minus					;- Taste
	rjmp	down						;Wert vermindern
	rjmp	mlop						;Keine Taste, noch eine Runde

next:
	cbr		flags,1<<enter				;Tastenflag lschen
	rjmp	sethi

up:	
	cbr		flags,1<<plus
	inc		druckmin
	mov		temp3,druckmin
	cpi		temp3,200
	brcs	up_ok1
	dec		druckmin
	rjmp	mlop
up_ok1:
	rcall	print_z
	rjmp	mlop

down:
	cbr		flags,1<<minus
	dec		druckmin
	mov		temp3,druckmin
	cpi		temp3,8
	brcc	down_ok1
	inc		druckmin
	rjmp	mlop
down_ok1:
	rcall	print_z
	rjmp	mlop

;--------------------------
sethi:
	cbr		flags,1<<enter|1<<DPunkt	;Flags Entertaste, Dezimalpunkt lschen
	rcall	scroll_high
	sbr		flags,1<<dpunkt				;Dezimalpunkt wieder ein
	mov		temp3,druckmax				;Unterer Schaltpunkt -> Display	

	cp		temp3,druckmin
	brcc	mm_ok
	mov		druckmax,druckmin
	inc		druckmax
	mov		temp3,druckmax
mm_ok:
	rcall	print_z						;Unterer Schaltpunkt -> Display	
	mov		temp2,druckmin
	inc		temp2
		
mlop1:
	sbrc	flags,enter					;Tasten abfragen
	rjmp	back
	sbrc	flags,plus
	rjmp	up1
	sbrc	flags,minus
	rjmp	down1
	rjmp	mlop1						;Keine Taste, noch eine Runde

back:
	cbr		flags,1<<enter
	sbr		flags,1<<max
	rcall	save
	rcall	clrd
	rcall	wait1
	rjmp	loop

up1:	
	cbr		flags,1<<plus
	inc		druckmax
	mov		temp3,druckmax
	cpi		temp3,240
	brcs	up_ok2
	dec		druckmax
	rjmp	mlop1
up_ok2:
	rcall	print_z
	rjmp	mlop1

down1:
	cbr		flags,1<<minus
	dec		druckmax
	mov		temp3,druckmax
	cp		druckmax,temp2
	brcc	down_ok2
	inc		druckmax
	rjmp	mlop1
down_ok2:
	rcall	print_z
	rjmp	mlop1
;-------------------------------------------
;EEprom
save:
	clr		temp1
	mov		temp2,druckmin
	rcall	WriteEE
	inc		temp1
	mov		temp2,druckmax
	rcall	WriteEE
	ret
load:
	clr		temp1
	rcall	ReadEE
	mov		druckmin,temp2
	inc		temp1
	rcall	ReadEE
	mov		druckmax,temp2
	ret

;------------------------------------
;EEprom	lesen Adr.temp1 Daten temp (1 Byte)
ReadEE:								;Adresse 0x01
nxtEE:	sbic	EECR,EEPE			;Bereit ?
		rjmp	nxtEE
		out		EEARL,temp1			;Adresse setzen
;		out		EEARH,null
		sbi		EECR,EERE			;ReadEnable in Controllregister
		in		temp2,EEDR			;Daten Holen
		ret
;------------------------------------
;EEprom	schreiben Adr.temp1 Daten temp (1 Byte)
WriteEE:
		cli
WriteEE1:
		sbic	EECR,EEPE			;Bereit ?
		rjmp	WriteEE1
		out		EEARL,temp1			;Adresse
;		out		EEARH,null
		out		EEDR,temp2			;Daten
		cli
		sbi		EECR,EEMPE			;Master Write Enable
		sbi		EECR,EEPE			;Write Enable
		sei
		ret

w04:
	ldi		tick,120
	rjmp	w1

wait1:
	rcall	wait
	rcall	wait
	rcall	wait
	rcall	wait
	rcall	wait


wait:
	ldi		tick,100
w1:	tst		tick
	brne	w1
	ret


clrd:
	ldi		temp1,blank
	mov		seg0,temp1
	mov		seg1,temp1
	mov		seg2,temp1
	ret

print_z:
	ldi		zh,high(bar*2)	;Zeiger auf 7Seg. Tabelle	
	ldi		zl,low(bar*2)
	add		zl,temp3			;*3 wg. Tabelle
	adc		zh,null
	add		zl,temp3
	adc		zh,null
	add		zl,temp3
	adc		zh,null
	lpm		seg0,z+				;7Seg von der Tab -> Display
	lpm		seg1,z+
	lpm		seg2,z+
	ret

scroll_low:
	rcall	clrd
	rcall	w04
	ldi		zh,high(setlow*2)	;Zeiger auf 7Seg. Tabelle	
	ldi		zl,low(setlow*2)
	rjmp	scroll
scroll_high:
	rcall	clrd
	rcall	w04
	ldi		zh,high(sethigh*2)	;Zeiger auf 7Seg. Tabelle	
	ldi		zl,low(sethigh*2)
	rjmp	scroll

scroll:
	lpm		temp1,z+
	tst		temp1
	breq	exscroll
	mov		seg0,seg1
	mov		seg1,seg2
	mov		seg2,temp1
	rcall	wait
	rjmp	scroll


exscroll:
	rcall	clrd
	ret

druck:	.include	"Drucktab.asm"
